unit hiEffectImg;

interface

uses
  Kol, Windows, Share, Debug;

type
  PByteArray = ^TByteArray;
  TByteArray = array[0..32767] of Byte;

  THIEffectImg = class(TDebug)
  private
    FBitmap: KOL.PBitmap;
  public
    _data_Level: THI_Event;
    _data_Bitmap: THI_Event;
    _event_onResult: THI_Event;
    _prop_Effects: byte;

    procedure Effect_Invert;
    procedure Effect_AddColorNoise(Amount: Integer);
    procedure Effect_AddMonoNoise(Amount: Integer);
    procedure Effect_AntiAlias;
    procedure Effect_Contrast(Amount: Integer);
    procedure Effect_FishEye(Amount: Integer);
    procedure Effect_GrayScale;
    procedure Effect_Lightness(Amount: Integer);
    procedure Effect_Darkness(Amount: Integer);
    procedure Effect_Saturation(Amount: Integer);
    procedure Effect_SplitBlur(Amount: Integer);
    procedure Effect_GaussianBlur(Amount: Integer);
    procedure Effect_Mosaic(Size: Integer);
    procedure Effect_Twist(Amount: Integer);
    procedure Effect_Splitlight(Amount: Integer);
    procedure Effect_Tile(Amount: Integer);
    procedure Effect_SpotLight(Amount: Integer; Spot: TRect);
    procedure Effect_Trace(Amount: Integer);
    procedure Effect_Emboss;
    procedure Effect_Solorize(Amount: Integer);
    procedure Effect_Posterize(Amount: Integer);

    procedure _work_doEffect(var _Data: TData; Index: word);
    procedure _var_Result(var _Data: TData; Index: word);
    procedure _var_EffectsList(var _Data: TData; Index: word);


    property Bitmap: KOL.PBitmap read FBitmap;// write FBitmap;

  end;

implementation

const
  Effects = 'GaussianBlur'+#13#10+'SplitBlur'+#13#10+'AddColorNoise'+#13#10+
            'AddMonoNoise'+#13#10+'Antialias'+#13#10+'Contrast'+#13#10+'Fisheye'+#13#10+
            'Lightness'+#13#10+'Darkness'+#13#10+'Saturation'+#13#10+'Mosaic'+#13#10+
            'Twist'+#13#10+'SplitLight'+#13#10+'Tile'+#13#10+'SpotLight'+#13#10+'Trace'+#13#10+
            'Emboss'+#13#10+'Solorize'+#13#10+'Posterize'+#13#10+'Grayscale'+#13#10+'Invert';


procedure THIEffectImg._work_doEffect;
var
  EffectsIndex,Level,i: integer;
  TmpBmp: KOL.PBitmap;
begin
  EffectsIndex := ToInteger(_Data);
  FBitmap.Free;
  FBitmap := NewBitmap(1,1);
  FBitmap.Assign(ToBitmapEvent(_data_Bitmap));
  FBitmap.PixelFormat := pf24bit;
  Level := ToIntegerEvent(_data_Level);
  case EffectsIndex of
     0: Effect_GaussianBlur   (Level);
     1: Effect_SplitBlur      (Level);
     2: Effect_AddColorNoise  (Level * 3);
     3: Effect_AddMonoNoise   (Level * 3);
     4: for i := 1 to Level do Effect_AntiAlias;
     5: Effect_Contrast       (Level * 3);
     6: Effect_FishEye        (Level div 10+1);
     7: Effect_Lightness      (Level * 2);
     8: Effect_Darkness       (Level * 2);
     9: Effect_Saturation     (255-((Level * 255) div 100));
    10: Effect_Mosaic         (Level div 2);
    11: Effect_Twist          (200-(Level * 2)+1);
    12: Effect_Splitlight     (Level div 20);
    13: Effect_Tile           (Level div 10);
    14: Effect_SpotLight      (Level, MakeRect (Level ,Level ,Level + Level*2, Level +Level*2));
    15: Effect_Trace          (Level div 10);
    16: for i := 1 to Level do  Effect_Emboss;
    17: Effect_Solorize       (255-((Level * 255) div 100));
    18: Effect_Posterize      (((Level * 255) div 100)+1);
    19: Effect_Grayscale;
    20: Effect_Invert;
  end;{Case}
  _hi_onEvent(_event_onResult,FBitmap);
end;


procedure THIEffectImg._var_Result;
begin
  dtBitmap(_Data,FBitmap);
end;

procedure THIEffectImg._var_EffectsList;
begin
  dtString(_Data,Effects);
end;

procedure PicInvert(src: KOL.PBitmap);
var w, h, x, y: Integer;
  p: PByteArray;
begin
  w := src.width;
  h := src.height;
  src.PixelFormat := pf24bit;
  for y := 0 to h - 1 do begin
    p := src.scanline[y];
    for x := 0 to w - 1 do begin
      p[x * 3] := not p[x * 3];
      p[x * 3 + 1] := not p[x * 3 + 1];
      p[x * 3 + 2] := not p[x * 3 + 2];
    end;
  end;
end;

function IntToByte(i: Integer): Byte;
begin
  if i > 255 then Result := 255
  else if i < 0 then Result := 0
  else Result := i;
end;

procedure AddColorNoise(var clip: KOL.PBitmap; Amount: Integer);
var
  p0: PByteArray;
  x, y, r, g, b: Integer;

begin
  for y := 0 to clip.Height - 1 do
  begin
    p0 := clip.ScanLine[y];
    for x := 0 to clip.Width - 1 do
    begin
      r := p0[x * 3] + (Random(Amount) - (Amount shr 1));
      g := p0[x * 3 + 1] + (Random(Amount) - (Amount shr 1));
      b := p0[x * 3 + 2] + (Random(Amount) - (Amount shr 1));
      p0[x * 3] := IntToByte(r);
      p0[x * 3 + 1] := IntToByte(g);
      p0[x * 3 + 2] := IntToByte(b);
    end;
  end;
end;

procedure AddMonoNoise(var clip: KOL.PBitmap; Amount: Integer);
var
  p0: PByteArray;
  x, y, a, r, g, b: Integer;
begin
  for y := 0 to clip.Height - 1 do
  begin
    p0 := clip.scanline[y];
    for x := 0 to clip.Width - 1 do
    begin
      a := Random(Amount) - (Amount shr 1);
      r := p0[x * 3] + a;
      g := p0[x * 3 + 1] + a;
      b := p0[x * 3 + 2] + a;
      p0[x * 3] := IntToByte(r);
      p0[x * 3 + 1] := IntToByte(g);
      p0[x * 3 + 2] := IntToByte(b);
    end;
  end;
end;

procedure AntiAliasRect(clip: KOL.PBitmap; XOrigin, YOrigin,
  XFinal, YFinal: Integer);
var Memo, x, y: Integer; (* Composantes primaires des points environnants *)
  p0, p1, p2: PByteArray;

begin
  if XFinal < XOrigin then begin Memo := XOrigin; XOrigin := XFinal; XFinal := Memo; end; (* Inversion des valeurs   *)
  if YFinal < YOrigin then begin Memo := YOrigin; YOrigin := YFinal; YFinal := Memo; end; (* si diffrence ngative*)
  XOrigin := max(1, XOrigin);
  YOrigin := max(1, YOrigin);
  XFinal := min(clip.width - 2, XFinal);
  YFinal := min(clip.height - 2, YFinal);
  clip.PixelFormat := pf24bit;
  for y := YOrigin to YFinal do begin
    p0 := clip.ScanLine[y - 1];
    p1 := clip.scanline[y];
    p2 := clip.ScanLine[y + 1];
    for x := XOrigin to XFinal do begin
      p1[x * 3] := (p0[x * 3] + p2[x * 3] + p1[(x - 1) * 3] + p1[(x + 1) * 3]) div 4;
      p1[x * 3 + 1] := (p0[x * 3 + 1] + p2[x * 3 + 1] + p1[(x - 1) * 3 + 1] + p1[(x + 1) * 3 + 1]) div 4;
      p1[x * 3 + 2] := (p0[x * 3 + 2] + p2[x * 3 + 2] + p1[(x - 1) * 3 + 2] + p1[(x + 1) * 3 + 2]) div 4;
    end;
  end;
end;

procedure AntiAlias(clip: KOL.PBitmap);
begin
  AntiAliasRect(clip, 0, 0, clip.width, clip.height);
end;

procedure Contrast(var clip: KOL.PBitmap; Amount: Integer);
var
  p0: PByteArray;
  rg, gg, bg, r, g, b, x, y: Integer;
begin
  for y := 0 to clip.Height - 1 do
  begin
    p0 := clip.scanline[y];
    for x := 0 to clip.Width - 1 do
    begin
      r := p0[x * 3];
      g := p0[x * 3 + 1];
      b := p0[x * 3 + 2];
      rg := (Abs(127 - r) * Amount) div 255;
      gg := (Abs(127 - g) * Amount) div 255;
      bg := (Abs(127 - b) * Amount) div 255;
      if r > 127 then r := r + rg else r := r - rg;
      if g > 127 then g := g + gg else g := g - gg;
      if b > 127 then b := b + bg else b := b - bg;
      p0[x * 3] := IntToByte(r);
      p0[x * 3 + 1] := IntToByte(g);
      p0[x * 3 + 2] := IntToByte(b);
    end;
  end;
end;

procedure FishEye(var Bmp, Dst: KOL.PBitmap; Amount: Extended);
var
  xmid, ymid: Single;
  fx, fy: Single;
  r1, r2: Single;
  ifx, ify: Integer;
  dx, dy: Single;
  rmax: Single;
  ty, tx: Integer;
  weight_x, weight_y: array[0..1] of Single;
  weight: Single;
  new_red, new_green: Integer;
  new_blue: Integer;
  total_red, total_green: Single;
  total_blue: Single;
  ix, iy: Integer;
  sli, slo: PByteArray;
begin
  xmid := Bmp.Width / 2;
  ymid := Bmp.Height / 2;
  rmax := Dst.Width * Amount;

  for ty := 0 to Dst.Height - 1 do begin
    for tx := 0 to Dst.Width - 1 do begin
      dx := tx - xmid;
      dy := ty - ymid;
      r1 := Sqrt(dx * dx + dy * dy);
      if r1 = 0 then begin
        fx := xmid;
        fy := ymid;
      end
      else begin
        r2 := rmax / 2 * (1 / (1 - r1 / rmax) - 1);
        fx := dx * r2 / r1 + xmid;
        fy := dy * r2 / r1 + ymid;
      end;
      ify := Trunc(fy);
      ifx := Trunc(fx);
                // Calculate the weights.
      if fy >= 0 then begin
        weight_y[1] := fy - ify;
        weight_y[0] := 1 - weight_y[1];
      end else begin
        weight_y[0] := -(fy - ify);
        weight_y[1] := 1 - weight_y[0];
      end;
      if fx >= 0 then begin
        weight_x[1] := fx - ifx;
        weight_x[0] := 1 - weight_x[1];
      end else begin
        weight_x[0] := -(fx - ifx);
        Weight_x[1] := 1 - weight_x[0];
      end;

      if ifx < 0 then
        ifx := Bmp.Width - 1 - (-ifx mod Bmp.Width)
      else if ifx > Bmp.Width - 1 then
        ifx := ifx mod Bmp.Width;
      if ify < 0 then
        ify := Bmp.Height - 1 - (-ify mod Bmp.Height)
      else if ify > Bmp.Height - 1 then
        ify := ify mod Bmp.Height;

      total_red := 0.0;
      total_green := 0.0;
      total_blue := 0.0;
      for ix := 0 to 1 do begin
        for iy := 0 to 1 do begin
          if ify + iy < Bmp.Height then
            sli := Bmp.scanline[ify + iy]
          else
            sli := Bmp.scanline[Bmp.Height - ify - iy];
          if ifx + ix < Bmp.Width then begin
            new_red := sli[(ifx + ix) * 3];
            new_green := sli[(ifx + ix) * 3 + 1];
            new_blue := sli[(ifx + ix) * 3 + 2];
          end
          else begin
            new_red := sli[(Bmp.Width - ifx - ix) * 3];
            new_green := sli[(Bmp.Width - ifx - ix) * 3 + 1];
            new_blue := sli[(Bmp.Width - ifx - ix) * 3 + 2];
          end;
          weight := weight_x[ix] * weight_y[iy];
          total_red := total_red + new_red * weight;
          total_green := total_green + new_green * weight;
          total_blue := total_blue + new_blue * weight;
        end;
      end;
      slo := Dst.scanline[ty];
      slo[tx * 3] := Round(total_red);
      slo[tx * 3 + 1] := Round(total_green);
      slo[tx * 3 + 2] := Round(total_blue);

    end;
  end;
end;

procedure GrayScale(var clip: KOL.PBitmap);
var
  p0: PByteArray;
  Gray, x, y: Integer;
begin
  for y := 0 to clip.Height - 1 do
  begin
    p0 := clip.scanline[y];
    for x := 0 to clip.Width - 1 do
    begin
      Gray := Round(p0[x * 3] * 0.3 + p0[x * 3 + 1] * 0.59 + p0[x * 3 + 2] * 0.11);
      p0[x * 3] := Gray;
      p0[x * 3 + 1] := Gray;
      p0[x * 3 + 2] := Gray;
    end;
  end;
end;


procedure Lightness(var clip: KOL.PBitmap; Amount: Integer);
var
  p0: PByteArray;
  r, g, b, x, y: Integer;
begin
  for y := 0 to clip.Height - 1 do begin
    p0 := clip.scanline[y];
    for x := 0 to clip.Width - 1 do
    begin
      r := p0[x * 3];
      g := p0[x * 3 + 1];
      b := p0[x * 3 + 2];
      p0[x * 3] := IntToByte(r + ((255 - r) * Amount) div 255);
      p0[x * 3 + 1] := IntToByte(g + ((255 - g) * Amount) div 255);
      p0[x * 3 + 2] := IntToByte(b + ((255 - b) * Amount) div 255);
    end;
  end;
end;

procedure Darkness(var src: KOL.PBitmap; Amount: Integer);
var
  p0: PByteArray;
  r, g, b, x, y: Integer;
begin
  src.pixelformat := pf24bit;
  for y := 0 to src.Height - 1 do begin
    p0 := src.scanline[y];
    for x := 0 to src.Width - 1 do
    begin
      r := p0[x * 3];
      g := p0[x * 3 + 1];
      b := p0[x * 3 + 2];
      p0[x * 3] := IntToByte(r - ((r) * Amount) div 255);
      p0[x * 3 + 1] := IntToByte(g - ((g) * Amount) div 255);
      p0[x * 3 + 2] := IntToByte(b - ((b) * Amount) div 255);
    end;
  end;
end;


procedure Saturation(var clip: KOL.PBitmap; Amount: Integer);
var
  p0: PByteArray;
  Gray, r, g, b, x, y: Integer;
begin
  for y := 0 to clip.Height - 1 do begin
    p0 := clip.scanline[y];
    for x := 0 to clip.Width - 1 do
    begin
      r := p0[x * 3];
      g := p0[x * 3 + 1];
      b := p0[x * 3 + 2];
      Gray := (r + g + b) div 3;
      p0[x * 3] := IntToByte(Gray + (((r - Gray) * Amount) div 255));
      p0[x * 3 + 1] := IntToByte(Gray + (((g - Gray) * Amount) div 255));
      p0[x * 3 + 2] := IntToByte(Gray + (((b - Gray) * Amount) div 255));
    end;
  end;
end;

procedure SmoothResize(var Src, Dst: KOL.PBitmap);
var
  x, y, xP, yP,
    yP2, xP2: Integer;
  Read, Read2: PByteArray;
  t, z, z2, iz2: Integer;
  pc: PBytearray;
  w1, w2, w3, w4: Integer;
  Col1r, col1g, col1b, Col2r, col2g, col2b: byte;
begin
  xP2 := ((src.Width - 1) shl 15) div Dst.Width;
  yP2 := ((src.Height - 1) shl 15) div Dst.Height;
  yP := 0;
  for y := 0 to Dst.Height - 1 do
  begin
    xP := 0;
    Read := src.ScanLine[yP shr 15];
    if yP shr 16 < src.Height - 1 then
      Read2 := src.ScanLine[yP shr 15 + 1]
    else
      Read2 := src.ScanLine[yP shr 15];
    pc := Dst.scanline[y];
    z2 := yP and $7FFF;
    iz2 := $8000 - z2;
    for x := 0 to Dst.Width - 1 do
    begin
      t := xP shr 15;
      Col1r := Read[t * 3];
      Col1g := Read[t * 3 + 1];
      Col1b := Read[t * 3 + 2];
      Col2r := Read2[t * 3];
      Col2g := Read2[t * 3 + 1];
      Col2b := Read2[t * 3 + 2];
      z := xP and $7FFF;
      w2 := (z * iz2) shr 15;
      w1 := iz2 - w2;
      w4 := (z * z2) shr 15;
      w3 := z2 - w4;
      pc[x * 3 + 2] :=
        (Col1b * w1 + Read[(t + 1) * 3 + 2] * w2 +
        Col2b * w3 + Read2[(t + 1) * 3 + 2] * w4) shr 15;
      pc[x * 3 + 1] :=
        (Col1g * w1 + Read[(t + 1) * 3 + 1] * w2 +
        Col2g * w3 + Read2[(t + 1) * 3 + 1] * w4) shr 15;
      pc[x * 3] :=
        (Col1r * w1 + Read2[(t + 1) * 3] * w2 +
        Col2r * w3 + Read2[(t + 1) * 3] * w4) shr 15;
      Inc(xP, xP2);
    end;
    Inc(yP, yP2);
  end;
end;

function TrimInt(i, Min, Max: Integer): Integer;
begin
  if i > Max then Result := Max
  else if i < Min then Result := Min
  else Result := i;
end;


procedure SmoothRotate(var Src, Dst: KOL.PBitmap; cx, cy: Integer;
  Angle: Extended);
type
  TFColor = record b, g, r: Byte end;
var
  Top,
    Bottom,
    Left,
    Right,
    eww, nsw,
    fx, fy,
    wx, wy: Extended;
  cAngle,
    sAngle: Double;
  xDiff,
    yDiff,
    ifx, ify,
    px, py,
    ix, iy,
    x, y: Integer;
  nw, ne,
    sw, se: TFColor;
  P1, P2, P3: Pbytearray;
begin
  Angle := angle;
  Angle := -Angle * Pi / 180;
  sAngle := Sin(Angle);
  cAngle := Cos(Angle);
  xDiff := (Dst.Width - Src.Width) div 2;
  yDiff := (Dst.Height - Src.Height) div 2;
  for y := 0 to Dst.Height - 1 do
  begin
    P3 := Dst.scanline[y];
    py := 2 * (y - cy) + 1;
    for x := 0 to Dst.Width - 1 do
    begin
      px := 2 * (x - cx) + 1;
      fx := (((px * cAngle - py * sAngle) - 1) / 2 + cx) - xDiff;
      fy := (((px * sAngle + py * cAngle) - 1) / 2 + cy) - yDiff;
      ifx := Round(fx);
      ify := Round(fy);

      if (ifx > -1) and (ifx < Src.Width) and (ify > -1) and (ify < Src.Height) then
      begin
        eww := fx - ifx;
        nsw := fy - ify;
        iy := TrimInt(ify + 1, 0, Src.Height - 1);
        ix := TrimInt(ifx + 1, 0, Src.Width - 1);
        P1 := Src.scanline[ify];
        P2 := Src.scanline[iy];
        nw.r := P1[ifx * 3];
        nw.g := P1[ifx * 3 + 1];
        nw.b := P1[ifx * 3 + 2];
        ne.r := P1[ix * 3];
        ne.g := P1[ix * 3 + 1];
        ne.b := P1[ix * 3 + 2];
        sw.r := P2[ifx * 3];
        sw.g := P2[ifx * 3 + 1];
        sw.b := P2[ifx * 3 + 2];
        se.r := P2[ix * 3];
        se.g := P2[ix * 3 + 1];
        se.b := P2[ix * 3 + 2];

        Top := nw.b + eww * (ne.b - nw.b);
        Bottom := sw.b + eww * (se.b - sw.b);
        P3[x * 3 + 2] := IntToByte(Round(Top + nsw * (Bottom - Top)));

        Top := nw.g + eww * (ne.g - nw.g);
        Bottom := sw.g + eww * (se.g - sw.g);
        P3[x * 3 + 1] := IntToByte(Round(Top + nsw * (Bottom - Top)));

        Top := nw.r + eww * (ne.r - nw.r);
        Bottom := sw.r + eww * (se.r - sw.r);
        P3[x * 3] := IntToByte(Round(Top + nsw * (Bottom - Top)));
      end;
    end;
  end;
end;


procedure SplitBlur(var clip: KOL.PBitmap; Amount: Integer);
var
  p0, p1, p2: PByteArray;
  cx, x, y: Integer;
  Buf: array[0..3, 0..2] of byte;
begin
  if Amount = 0 then Exit;
  for y := 0 to clip.Height - 1 do
  begin
    p0 := clip.scanline[y];
    if y - Amount < 0 then p1 := clip.scanline[y]
    else {y-Amount>0}  p1 := clip.ScanLine[y - Amount];
    if y + Amount < clip.Height then p2 := clip.ScanLine[y + Amount]
    else {y+Amount>=Height}  p2 := clip.ScanLine[clip.Height - y];

    for x := 0 to clip.Width - 1 do
    begin
      if x - Amount < 0 then cx := x
      else {x-Amount>0}  cx := x - Amount;
      Buf[0, 0] := p1[cx * 3];
      Buf[0, 1] := p1[cx * 3 + 1];
      Buf[0, 2] := p1[cx * 3 + 2];
      Buf[1, 0] := p2[cx * 3];
      Buf[1, 1] := p2[cx * 3 + 1];
      Buf[1, 2] := p2[cx * 3 + 2];
      if x + Amount < clip.Width then cx := x + Amount
      else {x+Amount>=Width}  cx := clip.Width - x;
      Buf[2, 0] := p1[cx * 3];
      Buf[2, 1] := p1[cx * 3 + 1];
      Buf[2, 2] := p1[cx * 3 + 2];
      Buf[3, 0] := p2[cx * 3];
      Buf[3, 1] := p2[cx * 3 + 1];
      Buf[3, 2] := p2[cx * 3 + 2];
      p0[x * 3] := (Buf[0, 0] + Buf[1, 0] + Buf[2, 0] + Buf[3, 0]) shr 2;
      p0[x * 3 + 1] := (Buf[0, 1] + Buf[1, 1] + Buf[2, 1] + Buf[3, 1]) shr 2;
      p0[x * 3 + 2] := (Buf[0, 2] + Buf[1, 2] + Buf[2, 2] + Buf[3, 2]) shr 2;
    end;
  end;
end;

procedure GaussianBlur(var clip: KOL.PBitmap; Amount: Integer);
var
  i: Integer;
begin
  for i := Amount downto 0 do
    SplitBlur(clip, 3);
end;

procedure Mosaic(var Bm: KOL.PBitmap; size: Integer);
var
  x, y, i, j: Integer;
  p1, p2: PByteArray;
  r, g, b: byte;
begin
  y := 0;
  repeat
    p1 := bm.scanline[y];
    x := 0;
    repeat
      j := 1;
      repeat
        p2 := bm.scanline[y];
        x := 0;
        repeat
          r := p1[x * 3];
          g := p1[x * 3 + 1];
          b := p1[x * 3 + 2];
          i := 1;
          repeat
            p2[x * 3] := r;
            p2[x * 3 + 1] := g;
            p2[x * 3 + 2] := b;
            inc(x);
            inc(i);
          until (x >= bm.width) or (i > size);
        until x >= bm.width;
        inc(j);
        inc(y);
      until (y >= bm.height) or (j > size);
    until (y >= bm.height) or (x >= bm.width);
  until y >= bm.height;
end;


procedure Twist(var Bmp, Dst: KOL.PBitmap; Amount: Integer);
var
  fxmid, fymid: Single;
  txmid, tymid: Single;
  fx, fy: Single;
  tx2, ty2: Single;
  r: Single;
  theta: Single;
  ifx, ify: Integer;
  dx, dy: Single;
  OFFSET: Single;
  ty, tx: Integer;
  weight_x, weight_y: array[0..1] of Single;
  weight: Single;
  new_red, new_green: Integer;
  new_blue: Integer;
  total_red, total_green: Single;
  total_blue: Single;
  ix, iy: Integer;
  sli, slo: PBytearray;

function ArcTan2(xt, yt: Single): Single;
begin
  if xt = 0 then
    if yt > 0 then
      Result := Pi / 2
    else
      Result := -(Pi / 2)
  else begin
    Result := ArcTan(yt / xt);
    if xt < 0 then
      Result := Pi + ArcTan(yt / xt);
  end;
end;

begin
  OFFSET := -(Pi / 2);
  dx := Bmp.Width - 1;
  dy := Bmp.Height - 1;
  r := Sqrt(dx * dx + dy * dy);
  tx2 := r;
  ty2 := r;
  txmid := (Bmp.Width - 1) / 2; //Adjust these to move center of rotation
  tymid := (Bmp.Height - 1) / 2; //Adjust these to move ......
  fxmid := (Bmp.Width - 1) / 2;
  fymid := (Bmp.Height - 1) / 2;
  if tx2 >= Bmp.Width then tx2 := Bmp.Width - 1;
  if ty2 >= Bmp.Height then ty2 := Bmp.Height - 1;

  for ty := 0 to Round(ty2) do begin
    for tx := 0 to Round(tx2) do begin
      dx := tx - txmid;
      dy := ty - tymid;
      r := Sqrt(dx * dx + dy * dy);
      if r = 0 then begin
        fx := 0;
        fy := 0;
      end
      else begin
        theta := ArcTan2(dx, dy) - r / Amount - OFFSET;
        fx := r * Cos(theta);
        fy := r * Sin(theta);
      end;
      fx := fx + fxmid;
      fy := fy + fymid;

      ify := Trunc(fy);
      ifx := Trunc(fx);
                // Calculate the weights.
      if fy >= 0 then begin
        weight_y[1] := fy - ify;
        weight_y[0] := 1 - weight_y[1];
      end else begin
        weight_y[0] := -(fy - ify);
        weight_y[1] := 1 - weight_y[0];
      end;
      if fx >= 0 then begin
        weight_x[1] := fx - ifx;
        weight_x[0] := 1 - weight_x[1];
      end else begin
        weight_x[0] := -(fx - ifx);
        Weight_x[1] := 1 - weight_x[0];
      end;

      if ifx < 0 then
        ifx := Bmp.Width - 1 - (-ifx mod Bmp.Width)
      else if ifx > Bmp.Width - 1 then
        ifx := ifx mod Bmp.Width;
      if ify < 0 then
        ify := Bmp.Height - 1 - (-ify mod Bmp.Height)
      else if ify > Bmp.Height - 1 then
        ify := ify mod Bmp.Height;

      total_red := 0.0;
      total_green := 0.0;
      total_blue := 0.0;
      for ix := 0 to 1 do begin
        for iy := 0 to 1 do begin
          if ify + iy < Bmp.Height then
            sli := Bmp.scanline[ify + iy]
          else
            sli := Bmp.scanline[Bmp.Height - ify - iy];
          if ifx + ix < Bmp.Width then begin
            new_red := sli[(ifx + ix) * 3];
            new_green := sli[(ifx + ix) * 3 + 1];
            new_blue := sli[(ifx + ix) * 3 + 2];
          end
          else begin
            new_red := sli[(Bmp.Width - ifx - ix) * 3];
            new_green := sli[(Bmp.Width - ifx - ix) * 3 + 1];
            new_blue := sli[(Bmp.Width - ifx - ix) * 3 + 2];
          end;
          weight := weight_x[ix] * weight_y[iy];
          total_red := total_red + new_red * weight;
          total_green := total_green + new_green * weight;
          total_blue := total_blue + new_blue * weight;
        end;
      end;
      slo := Dst.scanline[ty];
      slo[tx * 3] := Round(total_red);
      slo[tx * 3 + 1] := Round(total_green);
      slo[tx * 3 + 2] := Round(total_blue);
    end;
  end;
end;

procedure Splitlight(var clip: KOL.PBitmap; amount: Integer);
var x, y, i: Integer;
  p1: PByteArray;

function sinpixs(a: Integer): Integer;
begin
  result := variant(sin(a / 255 * pi / 2) * 255);
end;
begin
  for i := 1 to amount do
    for y := 0 to clip.height - 1 do begin
      p1 := clip.scanline[y];
      for x := 0 to clip.width - 1 do begin
        p1[x * 3] := sinpixs(p1[x * 3]);
        p1[x * 3 + 1] := sinpixs(p1[x * 3 + 1]);
        p1[x * 3 + 2] := sinpixs(p1[x * 3 + 2]);
      end;
    end;
end;


procedure Tile(src, dst: KOL.PBitmap; amount: Integer);
var w, h, w2, h2, i, j: Integer;
  bm: KOL.PBitmap;
begin
  w := src.width;
  h := src.height;
  dst.width := w;
  dst.height := h;
  //dst.Canvas.draw(0, 0, src);
  src.draw(dst.Canvas.Handle,0, 0);
  if (amount <= 0) or ((w div amount) < 5) or ((h div amount) < 5) then exit;
  h2 := h div amount;
  w2 := w div amount;
  bm := NewBitmap(1,1);
  bm.width := w2;
  bm.height := h2;
  bm.PixelFormat := pf24bit;
  smoothresize(src, bm);
  for j := 0 to amount - 1 do
    for i := 0 to amount - 1 do
      //dst.canvas.Draw(i * w2, j * h2, bm);
      bm.Draw(dst.Canvas.Handle, i * w2, j * h2);
  bm.free;
end;

procedure SpotLight(var src: KOL.PBitmap; Amount: Integer; Spot: TRect);
var bm: KOL.PBitmap;
  w, h: Integer;
begin
  Darkness(src, amount);
  w := src.Width;
  h := src.Height;
  bm := NewBitmap(1,1);
  bm.width := w;
  bm.height := h;
  bm.canvas.Brush.color := clblack;
  bm.canvas.FillRect(makerect(0, 0, w, h));
  bm.canvas.brush.Color := clwhite;
  bm.canvas.Ellipse(Spot.left, spot.top, spot.right, spot.bottom);
  bm.DrawTransparent(src.Canvas.Handle,0, 0,clwhite);
  bm.free;
end;


procedure Trace(src: KOL.PBitmap; intensity: Integer);
var
  x, y, i: Integer;
  P1, P2, P3, P4: PByteArray;
  tb, TraceB: byte;
  hasb: boolean;
  bitmap: KOL.PBitmap;
begin
  bitmap := NewBitmap(1,1);
  bitmap.width := src.width;
  bitmap.height := src.height;
  //bitmap.canvas.draw(0, 0, src);
  src.draw(bitmap.Canvas.Handle, 0, 0);
  bitmap.PixelFormat := pf8bit;
  src.PixelFormat := pf24bit;
  hasb := false;
  TraceB := $00;
  for i := 1 to Intensity do begin
    for y := 0 to BitMap.height - 2 do begin
      P1 := BitMap.ScanLine[y];
      P2 := BitMap.scanline[y + 1];
      P3 := src.scanline[y];
      P4 := src.scanline[y + 1];
      x := 0;
      repeat
        if p1[x] <> p1[x + 1] then begin
          if not hasb then begin
            tb := p1[x + 1];
            hasb := true;
            p3[x * 3] := TraceB;
            p3[x * 3 + 1] := TraceB;
            p3[x * 3 + 2] := TraceB;
          end
          else begin
            if p1[x] <> tb then
            begin
              p3[x * 3] := TraceB;
              p3[x * 3 + 1] := TraceB;
              p3[x * 3 + 2] := TraceB;
            end
            else
            begin
              p3[(x + 1) * 3] := TraceB;
              p3[(x + 1) * 3 + 1] := TraceB;
              p3[(x + 1) * 3 + 1] := TraceB;
            end;
          end;
        end;
        if p1[x] <> p2[x] then begin
          if not hasb then begin
            tb := p2[x];
            hasb := true;
            p3[x * 3] := TraceB;
            p3[x * 3 + 1] := TraceB;
            p3[x * 3 + 2] := TraceB;
          end
          else begin
            if p1[x] <> tb then
            begin
              p3[x * 3] := TraceB;
              p3[x * 3 + 1] := TraceB;
              p3[x * 3 + 2] := TraceB;
            end
            else
            begin
              p4[x * 3] := TraceB;
              p4[x * 3 + 1] := TraceB;
              p4[x * 3 + 2] := TraceB;
            end;
          end;
        end;
        inc(x);
      until x >= (BitMap.width - 2);
    end;
    if i > 1 then
      for y := BitMap.height - 1 downto 1 do begin
        P1 := BitMap.ScanLine[y];
        P2 := BitMap.scanline[y - 1];
        P3 := src.scanline[y];
        P4 := src.scanline[y - 1];
        x := Bitmap.width - 1;
        repeat
          if p1[x] <> p1[x - 1] then begin
            if not hasb then begin
              tb := p1[x - 1];
              hasb := true;
              p3[x * 3] := TraceB;
              p3[x * 3 + 1] := TraceB;
              p3[x * 3 + 2] := TraceB;
            end
            else begin
              if p1[x] <> tb then
              begin
                p3[x * 3] := TraceB;
                p3[x * 3 + 1] := TraceB;
                p3[x * 3 + 2] := TraceB;
              end
              else
              begin
                p3[(x - 1) * 3] := TraceB;
                p3[(x - 1) * 3 + 1] := TraceB;
                p3[(x - 1) * 3 + 2] := TraceB;
              end;
            end;
          end;
          if p1[x] <> p2[x] then begin
            if not hasb then begin
              tb := p2[x];
              hasb := true;
              p3[x * 3] := TraceB;
              p3[x * 3 + 1] := TraceB;
              p3[x * 3 + 2] := TraceB;
            end
            else begin
              if p1[x] <> tb then
              begin
                p3[x * 3] := TraceB;
                p3[x * 3 + 1] := TraceB;
                p3[x * 3 + 2] := TraceB;
              end
              else
              begin
                p4[x * 3] := TraceB;
                p4[x * 3 + 1] := TraceB;
                p4[x * 3 + 2] := TraceB;
              end;
            end;
          end;
          dec(x);
        until x <= 1;
      end;
  end;
  bitmap.free;
end;


procedure Emboss(var Bmp: KOL.PBitmap);
var
  x, y: Integer;
  p1, p2: Pbytearray;
begin
  for y := 0 to Bmp.Height - 2 do
  begin
    p1 := bmp.scanline[y];
    p2 := bmp.scanline[y + 1];
    for x := 0 to Bmp.Width - 4 do
    begin
      p1[x * 3] := (p1[x * 3] + (p2[(x + 3) * 3] xor $FF)) shr 1;
      p1[x * 3 + 1] := (p1[x * 3 + 1] + (p2[(x + 3) * 3 + 1] xor $FF)) shr 1;
      p1[x * 3 + 2] := (p1[x * 3 + 2] + (p2[(x + 3) * 3 + 2] xor $FF)) shr 1;
    end;
  end;

end;


procedure Solorize(src, dst: KOL.PBitmap; amount: Integer);
var w, h, x, y: Integer;
  ps, pd: PByteArray;
  c: Integer;
begin
  w := src.width;
  h := src.height;
  src.PixelFormat := pf24bit;
  dst.PixelFormat := pf24bit;
  for y := 0 to h - 1 do begin
    ps := src.scanline[y];
    pd := dst.scanline[y];
    for x := 0 to w - 1 do begin
      c := (ps[x * 3] + ps[x * 3 + 1] + ps[x * 3 + 2]) div 3;
      if c > amount then begin
        pd[x * 3] := 255 - ps[x * 3];
        pd[x * 3 + 1] := 255 - ps[x * 3 + 1];
        pd[x * 3 + 2] := 255 - ps[x * 3 + 2];
      end
      else begin
        pd[x * 3] := ps[x * 3];
        pd[x * 3 + 1] := ps[x * 3 + 1];
        pd[x * 3 + 2] := ps[x * 3 + 2];
      end;
    end;
  end;
end;

procedure Posterize(src, dst: KOL.PBitmap; amount: Integer);
var w, h, x, y: Integer;
  ps, pd: PByteArray;
  c: Integer;
begin
  w := src.width;
  h := src.height;
  src.PixelFormat := pf24bit;
  dst.PixelFormat := pf24bit;
  for y := 0 to h - 1 do begin
    ps := src.scanline[y];
    pd := dst.scanline[y];
    for x := 0 to w - 1 do begin
      pd[x * 3] := round(ps[x * 3] / amount) * amount;
      pd[x * 3 + 1] := round(ps[x * 3 + 1] / amount) * amount;
      pd[x * 3 + 2] := round(ps[x * 3 + 2] / amount) * amount;
    end;
  end;
end;


//----------------------
//-------------------------
//----------------------

procedure THIEffectImg.Effect_Invert;
begin
  PicInvert(Bitmap);
  //Invalidate;
end;

procedure THIEffectImg.Effect_AddColorNoise(Amount: Integer);
var BB: KOL.PBitmap;
begin
  BB := NewBitmap(1,1);
  BB.PixelFormat := pf24bit;
  BB.Assign(Bitmap);
  AddColorNoise(bb, Amount);
  Bitmap.Assign(BB);
  BB.Free;
end;

procedure THIEffectImg.Effect_AddMonoNoise(Amount: Integer);
var BB: KOL.PBitmap;
begin
  BB := NewBitmap(1,1);
  BB.PixelFormat := pf24bit;
  BB.Assign(Bitmap);
  AddMonoNoise(bb, Amount);
  Bitmap.Assign(BB);
  BB.Free;
end;

procedure THIEffectImg.Effect_AntiAlias;
begin
  AntiAlias(Bitmap);
  //Invalidate;
end;

procedure THIEffectImg.Effect_Contrast(Amount: Integer);
var BB: KOL.PBitmap;
begin
  BB := NewBitmap(1,1);
  BB.PixelFormat := pf24bit;
  BB.Assign(Bitmap);
  Contrast(bb, Amount);
  Bitmap.Assign(BB);
  BB.Free;
end;

procedure THIEffectImg.Effect_FishEye(Amount: Integer);
var BB1, BB2: KOL.PBitmap;
begin
  BB1 := NewBitmap(1,1);
  BB1.PixelFormat := pf24bit;
  BB1.Assign(Bitmap);
  BB2 := NewBitmap(1,1);
  BB2.PixelFormat := pf24bit;
  BB2.Assign(BB1);
  FishEye(BB1, BB2, Amount);
  Bitmap.Assign(BB2);
  BB1.Free;
  BB2.Free;
end;

procedure THIEffectImg.Effect_GrayScale;
var BB: KOL.PBitmap;
begin
  BB := NewBitmap(1,1);
  BB.PixelFormat := pf24bit;
  BB.Assign(Bitmap);
  GrayScale(BB);
  Bitmap.Assign(BB);
  BB.Free;
end;

procedure THIEffectImg.Effect_Lightness(Amount: Integer);
var BB: KOL.PBitmap;
begin
  BB := NewBitmap(1,1);
  BB.PixelFormat := pf24bit;
  BB.Assign(Bitmap);
  Lightness(BB, Amount);
  Bitmap.Assign(BB);
  BB.Free;
end;

procedure THIEffectImg.Effect_Darkness(Amount: Integer);
var BB: KOL.PBitmap;
begin
  BB := NewBitmap(1,1);
  BB.PixelFormat := pf24bit;
  BB.Assign(Bitmap);
  Darkness(BB, Amount);
  Bitmap.Assign(BB);
  BB.Free;
end;

procedure THIEffectImg.Effect_Saturation(Amount: Integer);
var BB: KOL.PBitmap;
begin
  BB := NewBitmap(1,1);
  BB.PixelFormat := pf24bit;
  BB.Assign(Bitmap);
  Saturation(BB, Amount);
  Bitmap.Assign(BB);
  BB.Free;
end;

procedure THIEffectImg.Effect_SplitBlur(Amount: Integer);
var BB: KOL.PBitmap;
begin
  BB := NewBitmap(1,1);
  BB.PixelFormat := pf24bit;
  BB.Assign(Bitmap);
  SplitBlur(BB, Amount);
  Bitmap.Assign(BB);
  BB.Free;
end;

procedure THIEffectImg.Effect_GaussianBlur(Amount: Integer);
var BB: KOL.PBitmap;
begin
  BB := NewBitmap(Bitmap.Width,Bitmap.Height);
  BB.PixelFormat := pf24bit;
  BB.Assign(Bitmap);
  GaussianBlur(BB, Amount);
  Bitmap.Assign(BB);
  BB.Free;
end;

procedure THIEffectImg.Effect_Mosaic(Size: Integer);
var BB: KOL.PBitmap;
begin
  BB := NewBitmap(1,1);
  BB.PixelFormat := pf24bit;
  BB.Assign(Bitmap);
  Mosaic(BB, Size);
  Bitmap.Assign(BB);
  BB.Free;
end;

procedure THIEffectImg.Effect_Twist(Amount: Integer);
var BB1, BB2: KOL.PBitmap;
begin
  BB1 := NewBitmap(1,1);
  BB1.PixelFormat := pf24bit;
  BB1.Assign(Bitmap);
  BB2 := NewBitmap(1,1);
  BB2.PixelFormat := pf24bit;
  BB2.Assign(BB1);
  Twist(BB1, BB2, Amount);
  Bitmap.Assign(BB2);
  BB1.Free;
  BB2.Free;
end;

procedure THIEffectImg.Effect_Trace(Amount: Integer);
var BB1, BB2: KOL.PBitmap;
begin
  BB1 := NewBitmap(1,1);
  BB1.PixelFormat := pf24bit;
  BB1.Assign(Bitmap);
  BB2 := NewBitmap(1,1);
  BB2.PixelFormat := pf24bit;
  BB2.Assign(BB1);
  Trace(BB2, Amount);
  Bitmap.Assign(BB2);
  BB1.Free;
  BB2.Free;
end;

procedure THIEffectImg.Effect_Splitlight(Amount: Integer);
var BB1: KOL.PBitmap;
begin
  BB1 := NewBitmap(1,1);
  BB1.PixelFormat := pf24bit;
  BB1.Assign(Bitmap);
  Splitlight(BB1, Amount);
  Bitmap.Assign(BB1);
  BB1.Free;
end;

procedure THIEffectImg.Effect_Tile(Amount: Integer);
var BB1, BB2: KOL.PBitmap;
begin
  BB1 := NewBitmap(1,1);
  BB1.PixelFormat := pf24bit;
  BB1.Assign(Bitmap);
  BB2 := NewBitmap(1,1);
  BB2.PixelFormat := pf24bit;
  BB2.Assign(BB1);
  Tile(BB1, BB2, Amount);
  Bitmap.Assign(BB2);
  BB1.Free;
  BB2.Free;
end;

procedure THIEffectImg.Effect_SpotLight(Amount: Integer; Spot: TRect);
var BB1, BB2: KOL.PBitmap;
begin
  BB1 := NewBitmap(1,1);
  BB1.PixelFormat := pf24bit;
  BB1.Assign(Bitmap);
  BB2 := NewBitmap(1,1);
  BB2.PixelFormat := pf24bit;
  BB2.Assign(BB1);
  SpotLight(BB2, Amount, Spot);
  Bitmap.Assign(BB2);
  BB1.Free;
  BB2.Free;
end;

procedure THIEffectImg.Effect_Emboss;
var BB1, BB2: KOL.PBitmap;
begin
  BB1 := NewBitmap(1,1);
  BB1.PixelFormat := pf24bit;
  BB1.Assign(Bitmap);
  BB2 := NewBitmap(1,1);
  BB2.PixelFormat := pf24bit;
  BB2.Assign(BB1);
  Emboss(BB2);
  Bitmap.Assign(BB2);
  BB1.Free;
  BB2.Free;
end;

procedure THIEffectImg.Effect_Solorize(Amount: Integer);
var BB1, BB2: KOL.PBitmap;
begin
  BB1 := NewBitmap(1,1);
  BB1.PixelFormat := pf24bit;
  BB1.Assign(Bitmap);
  BB2 := NewBitmap(1,1);
  BB2.PixelFormat := pf24bit;
  BB2.Assign(BB1);
  Solorize(BB1, BB2, Amount);
  Bitmap.Assign(BB2);
  BB1.Free;
  BB2.Free;
end;


procedure THIEffectImg.Effect_Posterize(Amount: Integer);
var BB1, BB2: KOL.PBitmap;
begin
  BB1 := NewBitmap(1,1);
  BB1.PixelFormat := pf24bit;
  BB1.Assign(Bitmap);
  BB2 := NewBitmap(1,1);
  BB2.PixelFormat := pf24bit;
  BB2.Assign(BB1);
  Posterize(BB1, BB2, Amount);
  Bitmap.Assign(BB2);
  BB1.Free;
  BB2.Free;
end;

end.
